from __future__ import annotations

from typing import List, Optional

from metricflow.errors.errors import ParsingException
from metricflow.model.objects.common import Metadata
from metricflow.model.objects.constraints.where import WhereClauseConstraint
from metricflow.model.objects.base import (
    HashableBaseModel,
    ModelWithMetadataParsing,
    PydanticCustomInputParser,
    PydanticParseableValueType,
)
from metricflow.object_utils import ExtendedEnum, hash_strings
from metricflow.references import MeasureReference
from metricflow.time.time_granularity import TimeGranularity
from metricflow.time.time_granularity import string_to_time_granularity


class MetricType(ExtendedEnum):
    """Currently supported metric types"""

    MEASURE_PROXY = "measure_proxy"
    RATIO = "ratio"
    EXPR = "expr"
    CUMULATIVE = "cumulative"


class MetricInputMeasure(PydanticCustomInputParser, HashableBaseModel):
    """Provides a pointer to a measure along with metric-specific processing directives

    If an alias is set, this will be used as the string name reference for this measure after the aggregation
    phase in the SQL plan.
    """

    name: str
    constraint: Optional[WhereClauseConstraint]
    alias: Optional[str]

    @classmethod
    def _from_yaml_value(cls, input: PydanticParseableValueType) -> MetricInputMeasure:
        """Parses a MetricInputMeasure from a string (name only) or object (struct spec) input

        For user input cases, the original YAML spec for a Metric included measure(s) specified as string names
        or lists of string names. As such, configs pre-dating the addition of this model type will only provide the
        base name for this object.
        """
        if isinstance(input, str):
            return MetricInputMeasure(name=input)
        else:
            raise ValueError(
                f"MetricInputMeasure inputs from model configs are expected to be of either type string or "
                f"object (key/value pairs), but got type {type(input)} with value: {input}"
            )

    @property
    def measure_reference(self) -> MeasureReference:
        """Property accessor to get the MeasureReference associated with this metric input measure"""
        return MeasureReference(element_name=self.name)

    @property
    def post_aggregation_measure_reference(self) -> MeasureReference:
        """Property accessor to get the MeasureReference with the aliased name, if appropriate"""
        return MeasureReference(element_name=self.alias or self.name)


class CumulativeMetricWindow(PydanticCustomInputParser, HashableBaseModel):
    """Describes the window of time the metric should be accumulated over. ie '1 day', '2 weeks', etc"""

    count: int
    granularity: TimeGranularity

    def to_string(self) -> str:  # noqa: D
        return f"{self.count} {self.granularity.value}"

    @classmethod
    def _from_yaml_value(cls, input: PydanticParseableValueType) -> CumulativeMetricWindow:
        """Parses a CumulativeMetricWindow from a string input found in a user provided model specification

        The CumulativeMetricWindow is always expected to be provided as a string in user-defined YAML configs.
        """
        if isinstance(input, str):
            return CumulativeMetricWindow.parse(input)
        else:
            raise ValueError(
                f"CumulativeMetricWindow inputs from model configs are expected to always be of type string, but got "
                f"type {type(input)} with value: {input}"
            )

    @staticmethod
    def parse(window: str) -> CumulativeMetricWindow:
        """Returns window values if parsing succeeds, None otherwise

        Output of the form: (<time unit count>, <time granularity>, <error message>) - error message is None if window is formatted properly
        """
        parts = window.split(" ")
        if len(parts) != 2:
            raise ParsingException(
                f"Invalid window ({window}) in cumulative metric. Should be of the form `<count> <granularity>`, ie `28 days`",
            )

        granularity = parts[1]
        # if we switched to python 3.9 this could just be `granularity = parts[0].removesuffix('s')
        if granularity.endswith("s"):
            # months -> month
            granularity = granularity[:-1]
        if granularity not in [item.value for item in TimeGranularity]:
            raise ParsingException(
                f"Invalid time granularity {granularity} in cumulative metric window string: ({window})",
            )

        count = parts[0]
        if not count.isdigit():
            raise ParsingException(f"Invalid count ({count}) in cumulative metric window string: ({window})")

        return CumulativeMetricWindow(
            count=int(count),
            granularity=string_to_time_granularity(granularity),
        )


class MetricTypeParams(HashableBaseModel):
    """Type params add additional context to certain metric types (the context depends on the metric type)"""

    measure: Optional[MetricInputMeasure]
    measures: Optional[List[MetricInputMeasure]]
    numerator: Optional[MetricInputMeasure]
    denominator: Optional[MetricInputMeasure]
    expr: Optional[str]
    window: Optional[CumulativeMetricWindow]
    grain_to_date: Optional[TimeGranularity]

    @property
    def numerator_measure_reference(self) -> Optional[MeasureReference]:
        """Return the measure reference, if any, associated with the metric input measure defined as the numerator"""
        return self.numerator.measure_reference if self.numerator else None

    @property
    def denominator_measure_reference(self) -> Optional[MeasureReference]:
        """Return the measure reference, if any, associated with the metric input measure defined as the denominator"""
        return self.denominator.measure_reference if self.denominator else None


class Metric(HashableBaseModel, ModelWithMetadataParsing):
    """Describes a metric"""

    name: str
    description: Optional[str]
    type: MetricType
    type_params: MetricTypeParams
    constraint: Optional[WhereClauseConstraint]
    metadata: Optional[Metadata]

    @property
    def input_measures(self) -> List[MetricInputMeasure]:
        """Return the complete list of input measure configurations for this metric"""
        tp = self.type_params
        res = tp.measures or []
        if tp.measure:
            res.append(tp.measure)
        if tp.numerator:
            res.append(tp.numerator)
        if tp.denominator:
            res.append(tp.denominator)

        return res

    @property
    def measure_references(self) -> List[MeasureReference]:
        """Return the measure references associated with all input measure configurations for this metric"""
        return [x.measure_reference for x in self.input_measures]

    @property
    def definition_hash(self) -> str:  # noqa: D
        values: List[str] = [self.name, self.type_params.expr or ""]
        if self.constraint:
            values.append(self.constraint.where)
            if self.constraint.linkable_names:
                values.extend(self.constraint.linkable_names)
        values.extend([m.element_name for m in self.measure_references])
        return hash_strings(values)



from metricflow.model.objects.user_configured_model import UserConfiguredModel


def test_model_serialization_deserialization(simple_user_configured_model: UserConfiguredModel) -> None:
    """Tests Pydantic serialization and deserialization of a UserConfiguredModel

    This ensures any custom parsing operations internal to our Pydantic models are properly applied to not only
    user-provided YAML input, but also to internal parsing operations based on serialized model objects.
    """
    serialized_model = simple_user_configured_model.json()
    deserialized_model = simple_user_configured_model.parse_raw(serialized_model)
    assert deserialized_model == simple_user_configured_model



from abc import ABC, abstractmethod

from metricflow.sql.sql_plan import SqlQueryPlanNode


class SqlQueryPlanOptimizer(ABC):
    """Optimize the SQL query plan in some way.

    e.g. a column pruner that removes unnecessary select columns in sub-queries.
    """

    @abstractmethod
    def optimize(self, node: SqlQueryPlanNode) -> SqlQueryPlanNode:  # noqa :D
        pass



from typing import List

from metricflow.instances import DataSourceElementReference
from metricflow.model.objects.data_source import DataSource
from metricflow.model.objects.elements.dimension import DimensionType
from metricflow.model.objects.user_configured_model import UserConfiguredModel
from metricflow.model.validations.validator_helpers import (
    DataSourceElementContext,
    DataSourceElementType,
    FileContext,
    ModelValidationRule,
    ValidationIssueType,
    validate_safely,
    ValidationError,
)
from metricflow.references import TimeDimensionReference


class AggregationTimeDimensionRule(ModelValidationRule):
    """Checks that the aggregation time dimension for a measure points to a valid time dimension in the data source."""

    @staticmethod
    @validate_safely(whats_being_done="checking aggregation time dimension for data sources in the model")
    def validate_model(model: UserConfiguredModel) -> List[ValidationIssueType]:  # noqa: D
        issues: List[ValidationIssueType] = []
        for data_source in model.data_sources:
            issues.extend(AggregationTimeDimensionRule._validate_data_source(data_source))

        return issues

    @staticmethod
    def _time_dimension_in_model(time_dimension_reference: TimeDimensionReference, data_source: DataSource) -> bool:
        for dimension in data_source.dimensions:
            if dimension.type == DimensionType.TIME and dimension.name == time_dimension_reference.element_name:
                return True
        return False

    @staticmethod
    @validate_safely(whats_being_done="checking aggregation time dimension for a data source")
    def _validate_data_source(data_source: DataSource) -> List[ValidationIssueType]:
        issues: List[ValidationIssueType] = []

        for measure in data_source.measures:
            measure_context = DataSourceElementContext(
                file_context=FileContext.from_metadata(metadata=data_source.metadata),
                data_source_element=DataSourceElementReference(
                    data_source_name=data_source.name, element_name=measure.name
                ),
                element_type=DataSourceElementType.MEASURE,
            )
            agg_time_dimension_reference = measure.checked_agg_time_dimension
            if not AggregationTimeDimensionRule._time_dimension_in_model(
                time_dimension_reference=agg_time_dimension_reference, data_source=data_source
            ):
                issues.append(
                    ValidationError(
                        context=measure_context,
                        message=f"In data source '{data_source.name}', measure '{measure.name}' has the aggregation "
                        f"time dimension set to '{agg_time_dimension_reference.element_name}', "
                        f"which is not a valid time dimension in the data source",
                    )
                )

        return issues



from __future__ import annotations
from copy import deepcopy

from dataclasses import dataclass, field
from functools import partial
from math import floor
from time import perf_counter
from typing import Callable, List, Optional, Tuple
from metricflow.dataflow.builder.source_node import SourceNodeBuilder
from metricflow.dataflow.dataflow_plan import BaseOutput, FilterElementsNode
from metricflow.dataset.convert_data_source import DataSourceToDataSetConverter
from metricflow.dataset.data_source_adapter import DataSourceDataSet

from metricflow.dataset.dataset import DataSet
from metricflow.engine.metricflow_engine import MetricFlowEngine, MetricFlowExplainResult, MetricFlowQueryRequest
from metricflow.instances import DataSourceElementReference, DataSourceReference, MetricModelReference
from metricflow.model.objects.data_source import DataSource
from metricflow.model.objects.elements.dimension import Dimension, DimensionType
from metricflow.model.objects.metric import Metric
from metricflow.model.objects.user_configured_model import UserConfiguredModel
from metricflow.model.semantic_model import SemanticModel
from metricflow.model.validations.validator_helpers import (
    DataSourceContext,
    DataSourceElementContext,
    DataSourceElementType,
    FileContext,
    MetricContext,
    ModelValidationResults,
    ValidationContext,
    ValidationError,
    ValidationIssue,
    ValidationWarning,
)
from metricflow.plan_conversion.column_resolver import DefaultColumnAssociationResolver
from metricflow.plan_conversion.dataflow_to_sql import DataflowToSqlQueryPlanConverter
from metricflow.plan_conversion.time_spine import TimeSpineSource
from metricflow.protocols.sql_client import SqlClient
from metricflow.specs import DimensionSpec, LinkableInstanceSpec
from metricflow.sql.sql_bind_parameters import SqlBindParameters


@dataclass
class QueryRenderingTools:
    """General tools that data warehosue validations use for rendering the validation queries

    This is necessary for the validation steps that generate raw partial queries against the individual data sources
    (e.g., selecting a single dimension column).
    """

    semantic_model: SemanticModel
    source_node_builder: SourceNodeBuilder
    converter: DataSourceToDataSetConverter
    time_spine_source: TimeSpineSource
    plan_converter: DataflowToSqlQueryPlanConverter

    def __init__(self, model: UserConfiguredModel, system_schema: str) -> None:  # noqa: D
        self.semantic_model = SemanticModel(user_configured_model=model)
        self.source_node_builder = SourceNodeBuilder(semantic_model=self.semantic_model)
        self.time_spine_source = TimeSpineSource(schema_name=system_schema)
        self.converter = DataSourceToDataSetConverter(
            column_association_resolver=DefaultColumnAssociationResolver(semantic_model=self.semantic_model)
        )
        self.plan_converter = DataflowToSqlQueryPlanConverter(
            column_association_resolver=DefaultColumnAssociationResolver(self.semantic_model),
            semantic_model=self.semantic_model,
            time_spine_source=self.time_spine_source,
        )


@dataclass
class DataWarehouseValidationTask:
    """Dataclass for defining a task to be used in the DataWarehouseModelValidator"""

    query_and_params_callable: Callable[[], Tuple[str, SqlBindParameters]]
    error_message: str
    context: Optional[ValidationContext] = None
    on_fail_subtasks: List[DataWarehouseValidationTask] = field(default_factory=lambda: [])


class DataWarehouseTaskBuilder:
    """Task builder for standard data warehouse validation tasks"""

    @staticmethod
    def _remove_identifier_link_specs(specs: Tuple[LinkableInstanceSpec, ...]) -> Tuple[LinkableInstanceSpec, ...]:
        """For the purposes of data warehouse validation, specs with identifier_links are unnecesary"""
        return tuple(spec for spec in specs if not spec.identifier_links)

    @staticmethod
    def _data_source_node(render_tools: QueryRenderingTools, data_source: DataSource) -> BaseOutput[DataSourceDataSet]:
        """Builds and returns the DataSourceDataSet node for the given data source"""
        data_source_semantics = render_tools.semantic_model.data_source_semantics.get_by_reference(
            DataSourceReference(data_source_name=data_source.name)
        )
        assert data_source_semantics

        source_nodes = render_tools.source_node_builder.create_from_data_sets(
            (render_tools.converter.create_sql_source_data_set(data_source_semantics),)
        )

        assert len(source_nodes) >= 1
        return source_nodes[0]

    @staticmethod
    def renderize(
        sql_client: SqlClient, plan_converter: DataflowToSqlQueryPlanConverter, plan_id: str, nodes: FilterElementsNode
    ) -> Tuple[str, SqlBindParameters]:
        """Generates a sql query plan and returns the rendered sql and execution_parameters"""
        sql_plan = plan_converter.convert_to_sql_query_plan(
            sql_engine_attributes=sql_client.sql_engine_attributes,
            sql_query_plan_id=plan_id,
            dataflow_plan_node=nodes,
        )

        rendered_plan = sql_client.sql_engine_attributes.sql_query_plan_renderer.render_sql_query_plan(sql_plan)
        return (rendered_plan.sql, rendered_plan.execution_parameters)

    @classmethod
    def gen_data_source_tasks(
        cls, model: UserConfiguredModel, sql_client: SqlClient, system_schema: str
    ) -> List[DataWarehouseValidationTask]:
        """Generates a list of tasks for validating the data sources of the model"""

        # we need a dimension to query that we know exists (i.e. the dimension
        # is guaranteed to not cause a problem) on each data source.
        # Additionally, we don't want to modify the original model, so we
        # first make a deep copy of it
        model = deepcopy(model)
        for data_source in model.data_sources:
            data_source.dimensions = list(data_source.dimensions) + [
                Dimension(name=f"validation_dim_for_{data_source.name}", type=DimensionType.CATEGORICAL, expr="1")
            ]

        render_tools = QueryRenderingTools(model=model, system_schema=system_schema)

        tasks: List[DataWarehouseValidationTask] = []
        for data_source in model.data_sources:
            source_node = cls._data_source_node(render_tools=render_tools, data_source=data_source)
            spec = DimensionSpec.from_name(name=f"validation_dim_for_{data_source.name}")
            filter_elements_node = FilterElementsNode(parent_node=source_node, include_specs=[spec])

            tasks.append(
                DataWarehouseValidationTask(
                    query_and_params_callable=partial(
                        cls.renderize,
                        sql_client=sql_client,
                        plan_converter=render_tools.plan_converter,
                        plan_id=f"{data_source.name}_validation",
                        nodes=filter_elements_node,
                    ),
                    context=DataSourceContext(
                        file_context=FileContext.from_metadata(metadata=data_source.metadata),
                        data_source=DataSourceReference(data_source_name=data_source.name),
                    ),
                    error_message=f"Unable to access data source `{data_source.name}` in data warehouse",
                )
            )

        return tasks

    @classmethod
    def gen_dimension_tasks(
        cls, model: UserConfiguredModel, sql_client: SqlClient, system_schema: str
    ) -> List[DataWarehouseValidationTask]:
        """Generates a list of tasks for validating the dimensions of the model

        The high level tasks returned are "short cut" queries which try to
        query all the dimensions for a given data source. If that query fails,
        one or more of the dimensions is incorrectly specified. Thus if the
        query fails, there are subtasks which query the individual dimensions
        on the data source to identify which have issues.
        """

        render_tools = QueryRenderingTools(model=model, system_schema=system_schema)

        tasks: List[DataWarehouseValidationTask] = []
        for data_source in model.data_sources:
            if not data_source.dimensions:
                continue

            source_node = cls._data_source_node(render_tools=render_tools, data_source=data_source)

            data_source_sub_tasks: List[DataWarehouseValidationTask] = []
            dataset = render_tools.converter.create_sql_source_data_set(data_source)
            data_source_specs = DataWarehouseTaskBuilder._remove_identifier_link_specs(
                dataset.instance_set.spec_set.dimension_specs + dataset.instance_set.spec_set.time_dimension_specs
            )
            for spec in data_source_specs:
                filter_elements_node = FilterElementsNode(parent_node=source_node, include_specs=[spec])
                data_source_sub_tasks.append(
                    DataWarehouseValidationTask(
                        query_and_params_callable=partial(
                            cls.renderize,
                            sql_client=sql_client,
                            plan_converter=render_tools.plan_converter,
                            plan_id=f"{data_source.name}_dim_{spec.element_name}_validation",
                            nodes=filter_elements_node,
                        ),
                        context=DataSourceElementContext(
                            file_context=FileContext.from_metadata(metadata=data_source.metadata),
                            data_source_element=DataSourceElementReference(
                                data_source_name=data_source.name, element_name=spec.element_name
                            ),
                            element_type=DataSourceElementType.DIMENSION,
                        ),
                        error_message=f"Unable to query dimension `{spec.element_name}` on data source `{data_source.name}` in data warehouse",
                    )
                )

            filter_elements_node = FilterElementsNode(parent_node=source_node, include_specs=data_source_specs)
            tasks.append(
                DataWarehouseValidationTask(
                    query_and_params_callable=partial(
                        cls.renderize,
                        sql_client=sql_client,
                        plan_converter=render_tools.plan_converter,
                        plan_id=f"{data_source.name}_all_dimensions_validation",
                        nodes=filter_elements_node,
                    ),
                    context=DataSourceContext(
                        file_context=FileContext.from_metadata(metadata=data_source.metadata),
                        data_source=DataSourceReference(data_source_name=data_source.name),
                    ),
                    error_message=f"Failed to query dimensions in data warehouse for data source `{data_source.name}`",
                    on_fail_subtasks=data_source_sub_tasks,
                )
            )
        return tasks

    @classmethod
    def gen_identifier_tasks(
        cls, model: UserConfiguredModel, sql_client: SqlClient, system_schema: str
    ) -> List[DataWarehouseValidationTask]:
        """Generates a list of tasks for validating the identifiers of the model

        The high level tasks returned are "short cut" queries which try to
        query all the identifiers for a given data source. If that query fails,
        one or more of the identifiers is incorrectly specified. Thus if the
        query fails, there are subtasks which query the individual identifiers
        on the data source to identify which have issues.
        """

        render_tools = QueryRenderingTools(model=model, system_schema=system_schema)

        tasks: List[DataWarehouseValidationTask] = []
        for data_source in model.data_sources:
            if not data_source.identifiers:
                continue
            source_node = cls._data_source_node(render_tools=render_tools, data_source=data_source)

            data_source_sub_tasks: List[DataWarehouseValidationTask] = []
            dataset = render_tools.converter.create_sql_source_data_set(data_source)
            data_source_specs = DataWarehouseTaskBuilder._remove_identifier_link_specs(
                dataset.instance_set.spec_set.identifier_specs
            )
            for spec in data_source_specs:
                filter_elements_node = FilterElementsNode(parent_node=source_node, include_specs=[spec])
                data_source_sub_tasks.append(
                    DataWarehouseValidationTask(
                        query_and_params_callable=partial(
                            cls.renderize,
                            sql_client=sql_client,
                            plan_converter=render_tools.plan_converter,
                            plan_id=f"{data_source.name}_identifier_{spec.element_name}_validation",
                            nodes=filter_elements_node,
                        ),
                        context=DataSourceElementContext(
                            file_context=FileContext.from_metadata(metadata=data_source.metadata),
                            data_source_element=DataSourceElementReference(
                                data_source_name=data_source.name, element_name=spec.element_name
                            ),
                            element_type=DataSourceElementType.IDENTIFIER,
                        ),
                        error_message=f"Unable to query identifier `{spec.element_name}` on data source `{data_source.name}` in data warehouse",
                    )
                )

            filter_elements_node = FilterElementsNode(parent_node=source_node, include_specs=data_source_specs)
            tasks.append(
                DataWarehouseValidationTask(
                    query_and_params_callable=partial(
                        cls.renderize,
                        sql_client=sql_client,
                        plan_converter=render_tools.plan_converter,
                        plan_id=f"{data_source.name}_all_identifiers_validation",
                        nodes=filter_elements_node,
                    ),
                    context=DataSourceContext(
                        file_context=FileContext.from_metadata(metadata=data_source.metadata),
                        data_source=DataSourceReference(data_source_name=data_source.name),
                    ),
                    error_message=f"Failed to query identifiers in data warehouse for data source `{data_source.name}`",
                    on_fail_subtasks=data_source_sub_tasks,
                )
            )
        return tasks

    @classmethod
    def gen_measure_tasks(
        cls, model: UserConfiguredModel, sql_client: SqlClient, system_schema: str
    ) -> List[DataWarehouseValidationTask]:
        """Generates a list of tasks for validating the measures of the model

        The high level tasks returned are "short cut" queries which try to
        query all the measures for a given data source. If that query fails,
        one or more of the measures is incorrectly specified. Thus if the
        query fails, there are subtasks which query the individual measures
        on the data source to identify which have issues.
        """

        render_tools = QueryRenderingTools(model=model, system_schema=system_schema)

        tasks: List[DataWarehouseValidationTask] = []
        for data_source in model.data_sources:
            if not data_source.measures:
                continue

            source_node = cls._data_source_node(render_tools=render_tools, data_source=data_source)

            data_source_sub_tasks: List[DataWarehouseValidationTask] = []
            dataset = render_tools.converter.create_sql_source_data_set(data_source)
            data_source_specs = dataset.instance_set.spec_set.measure_specs
            for spec in data_source_specs:
                filter_elements_node = FilterElementsNode(parent_node=source_node, include_specs=[spec])
                data_source_sub_tasks.append(
                    DataWarehouseValidationTask(
                        query_and_params_callable=partial(
                            cls.renderize,
                            sql_client=sql_client,
                            plan_converter=render_tools.plan_converter,
                            plan_id=f"{data_source.name}_measure_{spec.element_name}_validation",
                            nodes=filter_elements_node,
                        ),
                        context=DataSourceElementContext(
                            file_context=FileContext.from_metadata(metadata=data_source.metadata),
                            data_source_element=DataSourceElementReference(
                                data_source_name=data_source.name, element_name=spec.element_name
                            ),
                            element_type=DataSourceElementType.MEASURE,
                        ),
                        error_message=f"Unable to query measure `{spec.element_name}` on data source `{data_source.name}` in data warehouse",
                    )
                )

            filter_elements_node = FilterElementsNode(parent_node=source_node, include_specs=data_source_specs)
            tasks.append(
                DataWarehouseValidationTask(
                    query_and_params_callable=partial(
                        cls.renderize,
                        sql_client=sql_client,
                        plan_converter=render_tools.plan_converter,
                        plan_id=f"{data_source.name}_all_measures_validation",
                        nodes=filter_elements_node,
                    ),
                    context=DataSourceContext(
                        file_context=FileContext.from_metadata(metadata=data_source.metadata),
                        data_source=DataSourceReference(data_source_name=data_source.name),
                    ),
                    error_message=f"Failed to query measures in data warehouse for data source `{data_source.name}`",
                    on_fail_subtasks=data_source_sub_tasks,
                )
            )
        return tasks

    @staticmethod
    def _gen_metric_task_query_and_params(
        metric: Metric, mf_engine: MetricFlowEngine
    ) -> Tuple[str, SqlBindParameters]:  # noqa: D
        mf_query = MetricFlowQueryRequest.create_with_random_request_id(
            metric_names=[metric.name], group_by_names=[DataSet.metric_time_dimension_name()]
        )
        explain_result: MetricFlowExplainResult = mf_engine.explain(mf_request=mf_query)
        return (explain_result.rendered_sql.sql_query, explain_result.rendered_sql.bind_parameters)

    @classmethod
    def gen_metric_tasks(
        cls, model: UserConfiguredModel, sql_client: SqlClient, system_schema: str
    ) -> List[DataWarehouseValidationTask]:
        """Generates a list of tasks for validating the metrics of the model"""
        mf_engine = MetricFlowEngine(
            semantic_model=SemanticModel(user_configured_model=model),
            sql_client=sql_client,
            system_schema=system_schema,
        )
        tasks: List[DataWarehouseValidationTask] = []
        for metric in model.metrics:
            tasks.append(
                DataWarehouseValidationTask(
                    query_and_params_callable=partial(
                        cls._gen_metric_task_query_and_params,
                        metric=metric,
                        mf_engine=mf_engine,
                    ),
                    context=MetricContext(
                        file_context=FileContext.from_metadata(metadata=metric.metadata),
                        metric=MetricModelReference(metric_name=metric.name),
                    ),
                    error_message=f"Unable to query metric `{metric.name}`.",
                )
            )
        return tasks


class DataWarehouseModelValidator:
    """A Validator for checking specific tasks for the model against the Data Warehouse

    Data Warehouse Validations are validations that are done against the data
    warehouse based on the model configured by the user. Their purpose is to
    ensure that queries generated by MetricFlow won't fail when you go to use
    them (assuming the model has passed these validations before use).
    """

    def __init__(self, sql_client: SqlClient, system_schema: str) -> None:  # noqa: D
        self._sql_client = sql_client
        self._sql_schema = system_schema

    def run_tasks(
        self, tasks: List[DataWarehouseValidationTask], timeout: Optional[int] = None
    ) -> ModelValidationResults:
        """Runs the list of tasks as queries agains the data warehouse, returning any found issues

        Args:
            tasks: A list of tasks to run against the data warehouse
            timeout: An optional timeout. Default is None. When the timeout is hit, function will return early.

        Returns:
            A list of validation issues discovered when running the passed in tasks against the data warehosue
        """

        # Used for keeping track if we go past the max time
        start_time = perf_counter()

        issues: List[ValidationIssue] = []
        # TODO: Asyncio implementation
        for index, task in enumerate(tasks):
            if timeout is not None and perf_counter() - start_time > timeout:
                issues.append(
                    ValidationWarning(
                        context=None,
                        message=f"Hit timeout before completing all tasks. Completed {index}/{len(tasks)} tasks.",
                    )
                )
                break
            try:
                (query_string, query_params) = task.query_and_params_callable()
                self._sql_client.dry_run(stmt=query_string, sql_bind_parameters=query_params)
            except Exception as e:
                issues.append(
                    ValidationError(
                        context=task.context,
                        message=task.error_message + f"\nRecieved following error from data warehouse:\n{e}",
                    )
                )
                if task.on_fail_subtasks:
                    sub_task_timeout = floor(timeout - (perf_counter() - start_time)) if timeout else None
                    issues += self.run_tasks(tasks=task.on_fail_subtasks, timeout=sub_task_timeout).all_issues

        return ModelValidationResults.from_issues_sequence(issues)

    def validate_data_sources(
        self, model: UserConfiguredModel, timeout: Optional[int] = None
    ) -> ModelValidationResults:
        """Generates a list of tasks for validating the data sources of the model and then runs them

        Args:
            model: Model which to run data warehouse validations on
            timeout: An optional timeout. Default is None. When the timeout is hit, function will return early.

        Returns:
            A list of validation issues discovered when running the passed in tasks against the data warehosue
        """
        tasks = DataWarehouseTaskBuilder.gen_data_source_tasks(
            model=model, sql_client=self._sql_client, system_schema=self._sql_schema
        )
        return self.run_tasks(tasks=tasks, timeout=timeout)

    def validate_dimensions(self, model: UserConfiguredModel, timeout: Optional[int] = None) -> ModelValidationResults:
        """Generates a list of tasks for validating the dimensions of the model and then runs them

        Args:
            model: Model which to run data warehouse validations on
            timeout: An optional timeout. Default is None. When the timeout is hit, function will return early.

        Returns:
            A list of validation issues. If there are no validation issues, an empty list is returned.
        """
        tasks = DataWarehouseTaskBuilder.gen_dimension_tasks(
            model=model, sql_client=self._sql_client, system_schema=self._sql_schema
        )
        return self.run_tasks(tasks=tasks, timeout=timeout)

    def validate_identifiers(self, model: UserConfiguredModel, timeout: Optional[int] = None) -> ModelValidationResults:
        """Generates a list of tasks for validating the identifiers of the model and then runs them

        Args:
            model: Model which to run data warehouse validations on
            timeout: An optional timeout. Default is None. When the timeout is hit, function will return early.

        Returns:
            A list of validation issues. If there are no validation issues, an empty list is returned.
        """
        tasks = DataWarehouseTaskBuilder.gen_identifier_tasks(
            model=model, sql_client=self._sql_client, system_schema=self._sql_schema
        )
        return self.run_tasks(tasks=tasks, timeout=timeout)

    def validate_measures(self, model: UserConfiguredModel, timeout: Optional[int] = None) -> ModelValidationResults:
        """Generates a list of tasks for validating the measures of the model and then runs them

        Args:
            model: Model which to run data warehouse validations on
            timeout: An optional timeout. Default is None. When the timeout is hit, function will return early.

        Returns:
            A list of validation issues. If there are no validation issues, an empty list is returned.
        """
        tasks = DataWarehouseTaskBuilder.gen_measure_tasks(
            model=model, sql_client=self._sql_client, system_schema=self._sql_schema
        )
        return self.run_tasks(tasks=tasks, timeout=timeout)

    def validate_metrics(self, model: UserConfiguredModel, timeout: Optional[int] = None) -> ModelValidationResults:
        """Generates a list of tasks for validating the metrics of the model and then runs them

        Args:
            model: Model which to run data warehouse validations on
            timeout: An optional timeout. Default is None. When the timeout is hit, function will return early.

        Returns:
            A list of validation issues. If there are no validation issues, an empty list is returned.
        """

        tasks = DataWarehouseTaskBuilder.gen_metric_tasks(
            model=model, sql_client=self._sql_client, system_schema=self._sql_schema
        )
        return self.run_tasks(tasks=tasks, timeout=timeout)



from typing import List, Optional, Sequence

from metricflow.dataflow.sql_table import SqlTable
from metricflow.engine.models import Dimension
from metricflow.model.objects.common import FileSlice, Metadata
from metricflow.model.objects.constraints.where import WhereClauseConstraint
from metricflow.model.objects.data_source import DataSource, DataSourceOrigin, Mutability
from metricflow.model.objects.elements.identifier import Identifier
from metricflow.model.objects.elements.measure import Measure
from metricflow.model.objects.materialization import Materialization, MaterializationDestination
from metricflow.model.objects.metric import Metric, MetricType, MetricTypeParams


def default_meta() -> Metadata:
    """Returns a Metadata object with the required information"""

    return Metadata(
        repo_file_path="/not/from/a/repo",
        file_slice=FileSlice(
            filename="not_from_file.py",
            content="N/A",
            start_line_number=0,
            end_line_number=0,
        ),
    )


def materialization_with_guaranteed_meta(
    name: str,
    metrics: List[str],
    dimensions: List[str],
    description: str = "adhoc materialization",
    metadata: Metadata = default_meta(),
    destinations: List[MaterializationDestination] = [],
    destination_table: Optional[SqlTable] = None,
) -> Materialization:
    """Creates a materialization with the given input. If a metadata object is not supplied, a default metadata object is used"""

    return Materialization(
        name=name,
        description=description,
        metrics=metrics,
        dimensions=dimensions,
        destinations=destinations,
        destination_table=destination_table,
        metadata=metadata,
    )


def metric_with_guaranteed_meta(
    name: str,
    type: MetricType,
    type_params: MetricTypeParams,
    constraint: Optional[WhereClauseConstraint] = None,
    metadata: Metadata = default_meta(),
    description: str = "adhoc metric",
) -> Metric:
    """Creates a metric with the given input. If a metadata object is not supplied, a default metadata object is used"""

    return Metric(
        name=name,
        description=description,
        type=type,
        type_params=type_params,
        constraint=constraint,
        metadata=metadata,
    )


def data_source_with_guaranteed_meta(
    name: str,
    mutability: Mutability,
    description: Optional[str] = None,
    sql_table: Optional[str] = None,
    sql_query: Optional[str] = None,
    dbt_model: Optional[str] = None,
    metadata: Metadata = default_meta(),
    identifiers: Sequence[Identifier] = [],
    measures: Sequence[Measure] = [],
    dimensions: Sequence[Dimension] = [],
    origin: DataSourceOrigin = DataSourceOrigin.SOURCE,
) -> DataSource:
    """Creates a data source with the given input. If a metadata object is not supplied, a default metadata object is used"""

    return DataSource(
        name=name,
        mutability=mutability,
        description=description,
        sql_table=sql_table,
        sql_query=sql_query,
        dbt_model=dbt_model,
        identifiers=identifiers,
        measures=measures,
        dimensions=dimensions,
        origin=origin,
        metadata=metadata,
    )



from __future__ import annotations

import json
from typing import ClassVar, Optional, Sequence

import sqlalchemy

from metricflow.protocols.sql_client import SqlEngine, SqlEngineAttributes
from metricflow.sql.render.big_query import BigQuerySqlQueryPlanRenderer
from metricflow.sql.render.sql_plan_renderer import SqlQueryPlanRenderer
from metricflow.sql.sql_bind_parameters import SqlBindParameters
from metricflow.sql_clients.common_client import SqlDialect
from metricflow.sql_clients.sqlalchemy_dialect import SqlAlchemySqlClient


class BigQueryEngineAttributes:
    """Engine-specific attributes for the BigQuery query engine

    This is an implementation of the SqlEngineAttributes protocol for BigQuery
    """

    sql_engine_type: ClassVar[SqlEngine] = SqlEngine.BIGQUERY

    # SQL Engine capabilities
    date_trunc_supported: ClassVar[bool] = True
    full_outer_joins_supported: ClassVar[bool] = True
    indexes_supported: ClassVar[bool] = False
    multi_threading_supported: ClassVar[bool] = True
    timestamp_type_supported: ClassVar[bool] = True
    timestamp_to_string_comparison_supported: ClassVar[bool] = False
    # Cancelling should be possible, but not yet implemented.
    cancel_submitted_queries_supported: ClassVar[bool] = False

    # SQL Dialect replacement strings
    double_data_type_name: ClassVar[str] = "FLOAT64"
    timestamp_type_name: ClassVar[Optional[str]] = "DATETIME"

    # MetricFlow attributes
    sql_query_plan_renderer: ClassVar[SqlQueryPlanRenderer] = BigQuerySqlQueryPlanRenderer()


class BigQuerySqlClient(SqlAlchemySqlClient):
    """BigQuery implementation of SQL client"""

    @staticmethod
    def from_connection_details(url: str, password: Optional[str]) -> BigQuerySqlClient:  # noqa: D
        parsed_url = sqlalchemy.engine.url.make_url(url)
        dialect = SqlDialect.BIGQUERY.value
        if parsed_url.drivername != dialect:
            raise ValueError(f"Expected dialect '{dialect}' in {url}")

        if password is None:
            raise ValueError(f"Credentials not supplied for {url}")

        return BigQuerySqlClient(password=password)

    def __init__(self, project_id: str = "", password: Optional[str] = None, query_string: str = "") -> None:
        """Creates a new BigQueryDBClient and tags it for tracing as big query."""
        # Without pool_pre_ping, it's possible for timed-out connections to be returned to the client and cause errors.
        # However, this can cause increase latency for slow engines.
        self._password = password
        self._project_id = project_id
        super().__init__(
            engine=BigQuerySqlClient._create_bq_engine(
                query_string=query_string, project_id=project_id, password=password
            )
        )

    def _engine_specific_dry_run_implementation(self, stmt: str, bind_params: SqlBindParameters) -> None:
        """Overrides base `_engine_specific_dry_run_implementation` function for BigQuery specifics"""
        _engine = self._create_bq_engine(
            query_string="dry_run=true", project_id=self._project_id, password=self._password
        )
        with _engine.connect() as conn:
            conn.execute(sqlalchemy.text(stmt), bind_params.param_dict)

    @staticmethod
    def _create_bq_engine(
        query_string: str, project_id: str = "", password: Optional[str] = None
    ) -> sqlalchemy.engine.Engine:
        """Create the connection engine in SqlAlchemy to connect to BQ

        There are 2 methods of creating the BigQuery Engine,
        1. Using a service account JSON credential which will load into the credentials_info (Recommended for production use).
        2. Using ADC (Application Default Credentials) which allows a user to run `gcloud auth application-default login` to auth.
           This method would pass None to credentials_info and it will aggregate the credentials via,
           - Google's decision tree (https://google.aip.dev/auth/4110)

        Args:
            query_string: Additional query params to pass to engine construction.
            project_id: Project ID listed on the GCP platform.
            password: String representation of keyfile.json.
        """
        return sqlalchemy.create_engine(
            f"{SqlDialect.BIGQUERY.value}://{project_id}" + "/?" + query_string,
            credentials_info=json.loads(password) if password is not None else None,
            pool_size=10,
            max_overflow=10,
            pool_pre_ping=True,
        )

    @property
    def sql_engine_attributes(self) -> SqlEngineAttributes:
        """Collection of attributes and features specific to the BigQuery SQL engine"""
        return BigQueryEngineAttributes()

    def list_tables(self, schema_name: str) -> Sequence[str]:  # noqa: D
        with self.engine_connection(engine=self._engine) as conn:
            insp = sqlalchemy.inspection.inspect(conn)
            schema_dot_tables = insp.get_table_names(schema=schema_name)
            return [x.replace(schema_name + ".", "") for x in schema_dot_tables]

    def cancel_submitted_queries(self) -> None:  # noqa: D
        raise NotImplementedError



from metricflow.model.model_transformer import ModelTransformer
from metricflow.model.objects.user_configured_model import UserConfiguredModel
from metricflow.model.transformations.transform_rule import ModelTransformRule


class SliceNamesRule(ModelTransformRule):
    """Slice the names of data source elements in a model

    NOTE: specifically for testing
    """

    @staticmethod
    def transform_model(model: UserConfiguredModel) -> UserConfiguredModel:  # noqa: D
        for data_source in model.data_sources:
            data_source.name = data_source.name[:3]
        return model


def test_can_configure_model_transform_rules(simple_model__pre_transforms: UserConfiguredModel) -> None:  # noqa: D
    pre_model = simple_model__pre_transforms
    assert not all(len(x.name) == 3 for x in pre_model.data_sources)

    # Confirms that a custom transformation works for pre-validation transform
    pre_model = ModelTransformer.pre_validation_transform_model(pre_model, rules=[SliceNamesRule()])
    assert all(len(x.name) == 3 for x in pre_model.data_sources)

    post_model = simple_model__pre_transforms
    assert not all(len(x.name) == 3 for x in post_model.data_sources)

    # Confirms that a custom transformation works for post-validation transform
    post_model = ModelTransformer.post_validation_transform_model(post_model, rules=[SliceNamesRule()])
    assert all(len(x.name) == 3 for x in post_model.data_sources)



from _pytest.fixtures import FixtureRequest
from typing import List
from metricflow.dataflow.sql_table import SqlTable
from metricflow.protocols.sql_client import SqlClient
from metricflow.sql.sql_exprs import (
    SqlCastToTimestampExpression,
    SqlStringLiteralExpression,
)
from metricflow.sql.sql_plan import (
    SqlSelectColumn,
    SqlTableFromClauseNode,
    SqlSelectStatementNode,
    SqlOrderByDescription,
    SqlJoinDescription,
)
from metricflow.test.fixtures.setup_fixtures import MetricFlowTestSessionState
from metricflow.test.sql.compare_sql_plan import assert_rendered_sql_equal


def test_cast_to_timestamp(
    request: FixtureRequest,
    mf_test_session_state: MetricFlowTestSessionState,
    sql_client: SqlClient,
) -> None:
    """Tests rendering of the cast to timestamp expression in a query."""

    select_columns = [
        SqlSelectColumn(
            expr=SqlCastToTimestampExpression(
                arg=SqlStringLiteralExpression(
                    literal_value="2020-01-01",
                )
            ),
            column_alias="col0",
        ),
    ]

    from_source = SqlTableFromClauseNode(sql_table=SqlTable(schema_name="foo", table_name="bar"))

    from_source = from_source
    from_source_alias = "a"
    joins_descs: List[SqlJoinDescription] = []
    where = None
    group_bys: List[SqlSelectColumn] = []
    order_bys: List[SqlOrderByDescription] = []

    assert_rendered_sql_equal(
        request=request,
        mf_test_session_state=mf_test_session_state,
        select_node=SqlSelectStatementNode(
            description="Test Cast to Timestamp Expression",
            select_columns=tuple(select_columns),
            from_source=from_source,
            from_source_alias=from_source_alias,
            joins_descs=tuple(joins_descs),
            where=where,
            group_bys=tuple(group_bys),
            order_bys=tuple(order_bys),
        ),
        plan_id="plan0",
        sql_client=sql_client,
    )



from metricflow.inference.models import InferenceSignalType


def test_inference_type_node_conflict() -> None:
    """Make sure the inference signal type hierarchy is correctly configured."""

    # IDENTIFIER
    assert InferenceSignalType.ID.UNKNOWN.is_subtype_of(InferenceSignalType.UNKNOWN)
    assert not InferenceSignalType.ID.UNKNOWN.is_subtype_of(InferenceSignalType.DIMENSION.UNKNOWN)
    assert not InferenceSignalType.ID.UNKNOWN.is_subtype_of(InferenceSignalType.MEASURE.UNKNOWN)

    assert InferenceSignalType.ID.UNIQUE.is_subtype_of(InferenceSignalType.ID.UNKNOWN)
    assert InferenceSignalType.ID.FOREIGN.is_subtype_of(InferenceSignalType.ID.UNKNOWN)
    assert InferenceSignalType.ID.PRIMARY.is_subtype_of(InferenceSignalType.ID.UNIQUE)

    # DIMENSION
    assert InferenceSignalType.DIMENSION.UNKNOWN.is_subtype_of(InferenceSignalType.UNKNOWN)
    assert not InferenceSignalType.DIMENSION.UNKNOWN.is_subtype_of(InferenceSignalType.ID.UNKNOWN)
    assert not InferenceSignalType.DIMENSION.UNKNOWN.is_subtype_of(InferenceSignalType.MEASURE.UNKNOWN)

    assert InferenceSignalType.DIMENSION.CATEGORICAL.is_subtype_of(InferenceSignalType.DIMENSION.UNKNOWN)
    assert InferenceSignalType.DIMENSION.TIME.is_subtype_of(InferenceSignalType.DIMENSION.UNKNOWN)

    # MEASURE
    assert InferenceSignalType.MEASURE.UNKNOWN.is_subtype_of(InferenceSignalType.UNKNOWN)
    assert not InferenceSignalType.MEASURE.UNKNOWN.is_subtype_of(InferenceSignalType.ID.UNKNOWN)
    assert not InferenceSignalType.MEASURE.UNKNOWN.is_subtype_of(InferenceSignalType.DIMENSION.UNKNOWN)



"""This module contains custom helper click types."""

from typing import Any, Callable, Dict, Generic, List, Optional, Sequence, TypeVar, Tuple
import click

T = TypeVar("T")


# NOTE: unfortunately python 3.9 has no way of properly typing args and kwargs since it has no
# typing.ParamSpec. Since click uses args and kwargs, we have to add type ignore.
#
# Also, some click interfaces require us to return explicit Any.


class SequenceParamType(click.ParamType, Generic[T]):
    """A click parameter that is a list of elements.

    It can be used to parse strings like "1,2,4,8" into a Sequence[int], for example.
    """

    name = "sequence"

    def __init__(  # type: ignore
        self,
        value_converter: Callable[[str], T] = lambda x: x,  # type: ignore
        min_length: int = 0,
        max_length: Optional[int] = None,
        separator: str = ",",
        *args,
        **kwargs,
    ) -> None:
        """Initialize the sequence param type.

        value_converter: a function to convert each list value.
            Defaults to not converting and keeping them as strings.
        min_length: minimum length for the list, inclusive.
            Defaults to 0.
        max_length: maximum length for the list, inclusive.
            If None, there is no maximum length.
            Defaults to None.
        separator: the list separator.
            Defaults to `,`.
        """
        self.value_converter = value_converter
        self.min_length = min_length
        self.max_length = max_length
        self.separator = separator

        super().__init__(*args, **kwargs)

    def convert(  # noqa: D
        self, value: str, param: Optional[click.Parameter], ctx: Optional[click.Context]
    ) -> Sequence[T]:
        if len(value) == 0:
            return []

        str_values = value.split(self.separator)

        if len(str_values) < self.min_length:
            self.fail(
                f"list has length {len(str_values)}, which is less than the minimum allowed length of {self.min_length}",
                param,
                ctx,
            )

        if self.max_length is not None and len(str_values) > self.max_length:
            self.fail(
                f"list has length {len(str_values)}, which is more than the maximum allowed length of {self.max_length}",
                param,
                ctx,
            )

        converted_values: List[T] = []
        for str_val in str_values:
            try:
                converted_val = self.value_converter(str_val)
            except Exception:
                self.fail(f"{str_val} could not be converted by value converter", param, ctx)
            converted_values.append(converted_val)

        return converted_values


class MutuallyExclusiveOption(click.Option):
    """Type for making click options mutually exclusive.

    Usage:
    ```
    @option(
        '--arg-1',
        cls=MutuallyExclusiveOption,
        mutually_exclusive=["arg_2"],
        # ...
    )
    @option(
        '--arg-2',
        cls=MutuallyExclusiveOption,
        mutually_exclusive=["arg_1"],
        # ...
    )
    ```
    """

    def __init__(  # type: ignore
        self,
        *args,
        **kwargs,
    ):
        """Initialize the option.

        mutually_exclusive: A list of strings that indicate incompatible options.
        """

        self.mutually_exclusive = frozenset(kwargs.pop("mutually_exclusive", []))
        if len(self.mutually_exclusive) > 0:
            exclude_str = ",".join(self.mutually_exclusive)
            kwargs["help"] += f" [mutually_exclusive({exclude_str})]"

        super(MutuallyExclusiveOption, self).__init__(*args, **kwargs)

    def handle_parse_result(  # type: ignore # noqa: D
        self, ctx: click.Context, opts: Dict[str, Any], args: List[str]
    ) -> Tuple[Any, List[str]]:
        mutually_exclusive_opts_present = len(self.mutually_exclusive.intersection(opts)) > 0

        if mutually_exclusive_opts_present and self.name in opts:
            exclude_str = ",".join(f"'--{opt}'" for opt in self.mutually_exclusive)
            raise click.BadOptionUsage(
                self.name, f"cannot use option '--{self.name}' because it is mutually exclusive with {exclude_str}", ctx
            )

        return super(MutuallyExclusiveOption, self).handle_parse_result(ctx, opts, args)
